==> bogofilter-0.16.0.tar.gz: bogofilter-0.16.0/src/robinson.c <==
/* $Id: robinson.c,v 1.67 2004/01/01 13:34:50 relson Exp $ */

/*****************************************************************************

NAME:
   robinson.c -- implements f(w) and S, or Fisher, algorithm for computing spamicity.

******************************************************************************/

#include "common.h"

#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "bogoconfig.h"
#include "bogofilter.h"
#include "collect.h"
#include "datastore.h"
#ifdef	ENABLE_DEPRECATED_CODE
#include "degen.h"
#endif
#include "msgcounts.h"
#include "prob.h"
#include "robinson.h"
#include "rstats.h"
#include "wordhash.h"
#include "wordlists.h"

extern int Rtable;

double	thresh_rtable = 0.0;		/* used in fisher.c */

static rob_stats_t  rob_stats;

const parm_desc rob_parm_table[] =	/* needed by fisher.c */
{
    { "robx",		  CP_DOUBLE,	{ (void *) &robx } },
    { "robs",		  CP_DOUBLE,	{ (void *) &robs } },
    { "thresh_rtable",	  CP_DOUBLE,	{ (void *) &thresh_rtable } },
    { NULL,		  CP_NONE,	{ (void *) NULL } },
};

/* Function Prototypes */

static void	rob_initialize_constants(void);
static double	rob_get_spamicity(size_t robn, FLOAT P, FLOAT Q);
static void	rob_print_summary(void);

/* Static Variables */

rf_method_t rf_robinson_method = {	/* needed by config.c */
    {
	"robinson",			/* const char		  *name;		*/
	rob_parm_table,	 		/* m_parm_table		  *parm_table		*/
	rob_initialize_constants,	/* m_initialize_constants *initialize_constants	*/
	rob_compute_spamicity, 		/* m_compute_spamicity	  *compute_spamicity	*/
	mth_spamicity,			/* m_spamicity		  *spamicity		*/
	mth_status,			/* m_status		  *status		*/
	rob_print_stats, 		/* m_print_bogostats	  *print_stats		*/
	rob_cleanup, 			/* m_free		  *cleanup		*/
    },
    rob_get_spamicity,			/* rf_get_spamicity	  *get_spamicity	*/
    rob_print_summary			/* rf_print_summary	  *print_summary	*/
};

/* Function Definitions */

void rob_print_stats(FILE *fp)
{
    bool unsure = unsure_stats && ((*method->status)() == RC_UNSURE) && verbose;

    fp = NULL; 	/* quench compiler warning */
    if (Rtable || unsure || verbose >= 2 ||
	rob_stats.s.spamicity > thresh_stats ||
	rob_stats.s.spamicity > thresh_rtable)
	rstats_print(unsure);
}

static void wordprob_add(wordcnts_t *cnts, uint count, uint bad)
{
    if (bad)
	cnts->bad += count;
    else
	cnts->good += count;
}

static double wordprob_result(wordcnts_t *cnt)
{
    double fw = calc_prob(cnt->good, cnt->bad);
    return fw;
}

static void lookup(const word_t *token, wordcnts_t *cnts)
{
    int override=0;
    wordlist_t* list;

    if (fBogotune) {
	wordprop_t *wp = wordhash_search_memory(token);
	if (wp) {
	    cnts->good = wp->cnts.good;
	    cnts->bad  = wp->cnts.bad;
	}
	return;
    }

    for (list=word_lists; list != NULL; list=list->next)
    {
	size_t i;
	dsv_t val;

	if (override > list->override)	/* if already found */
	    break;

	if (ds_read(list->dsh, token, &val) != 0)
	    continue;			/* not found */

	if (list->ignore)		/* if on ignore list */
	    return;

	override=list->override;

	for (i=0; i<COUNTOF(val.count); i++) {
	    /* Protect against negatives */
	    if ((int) val.count[i] < 0) {
		val.count[i] = 0;
		ds_write(list->dsh, token, &val);
	    }
	    wordprob_add(cnts, val.count[i], list->bad[i]);
	}

	if (DEBUG_ROBINSON(1)) {
	    fprintf(dbgout, "%2d %2d \n", (int) cnts->good, (int) cnts->bad);
	    word_puts(token, 0, dbgout);
	    fputc('\n', dbgout);
	}
    }

    return;
}

double lookup_and_score(const word_t *token, wordcnts_t *cnts)
{
    double prob;

#ifdef	ENABLE_DEPRECATED_CODE
    const char *colon;
#endif

    if (cnts->bad == 0 && cnts->good == 0)
	lookup(token, cnts);

#ifdef	ENABLE_DEPRECATED_CODE
    if (header_degen && cnts->bad == 0 && cnts->good == 0 && (colon = strchr((const char *)(token->text), ':')) != NULL) {
	word_t *tword = word_new((unsigned const char *)(colon+1), strlen(colon+1));
	wordcnts_t tcnts;
	wordcnts_init(&tcnts);
	lookup_and_score(tword, &tcnts);
	wordcnts_incr(cnts, &tcnts);
	prob = wordprob_result(cnts);
	word_free(tword);
    }
    else
    if (cnts->bad != 0 || cnts->good != 0) {
	prob = wordprob_result(cnts);
    }
    else
    if (degen_enabled)
    {
	degen_enabled = false;	/* Disable further recursion */
	prob = degen(token, cnts);
	degen_enabled = true;	/* Enable further recursion */
    }
    else
#endif
	prob = wordprob_result(cnts);

    return prob;
}

static double compute_probability(const word_t *token, wordcnts_t *cnts)
{
    double prob;

    if (cnts->bad != 0 || cnts->good != 0 || msg_count_file)
	/* A msg-count file already has the values needed */
	prob = wordprob_result(cnts);
    else
	/* Otherwise lookup the word and get its score */
	prob = lookup_and_score(token, cnts);

    return prob;
}

double rob_compute_spamicity(wordhash_t *wh, FILE *fp) /*@globals errno@*/
/* selects the best spam/non-spam indicators and calculates Robinson's S */
{
    hashnode_t *node;

    FLOAT P = {1.0, 0};		/* Robinson's P */
    FLOAT Q = {1.0, 0};		/* Robinson's Q */

    double spamicity;
    size_t robn = 0;
    size_t count = 0;
    bool need_stats;

    (void) fp; 	/* quench compiler warning */

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "### rob_compute_spamicity() begins\n");

    Rtable |= verbose > 3;
    need_stats = Rtable || verbose || passthrough;

    if (need_stats)
	rstats_init();

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "min_dev: %f, robs: %f, robx: %f\n", 
				   min_dev, robs, robx);

    wordhash_sort(wh);

    for(node = wordhash_first(wh); node != NULL; node = wordhash_next(wh))
    {
	double prob;
	word_t *token;
	wordcnts_t *cnts;
	wordprop_t *props;

	if (!fBogotune) {
	    props = (wordprop_t *) node->buf;
	    cnts  = &props->cnts;
	    token = node->key;
	}
	else {
	    cnts = (wordcnts_t *) node;
	    token = NULL;
	}

	count += 1;

	prob = compute_probability(token, cnts);

	if (need_stats)
	    rstats_add(token, prob, cnts);

	/* Robinson's P and Q; accumulation step */
        /*
	 * P = 1 - ((1-p1)*(1-p2)*...*(1-pn))^(1/n)     [spamminess]
         * Q = 1 - (p1*p2*...*pn)^(1/n)                 [non-spamminess]
	 */
        if (fabs(EVEN_ODDS - prob) - min_dev >= EPS) {
	    int e;

	    P.mant *= 1-prob;
	    if (P.mant < 1.0e-200) {
		P.mant = frexp(P.mant, &e);
		P.exp += e;
	    }

	    Q.mant *= prob;
	    if (Q.mant < 1.0e-200) {
		Q.mant = frexp(Q.mant, &e);
		Q.exp += e;
	    }
            robn ++;
        }

	if (DEBUG_ROBINSON(3)) {
	    (void)fprintf(dbgout, "%3lu %3lu %f ",
			  (unsigned long)robn, (unsigned long)count, prob);
	    (void)word_puts(token, 0, dbgout);
	    (void)fputc('\n', dbgout);
	}
    }

    /* Robinson's P, Q and S
    ** S = (P - Q) / (P + Q)                        [combined indicator]
    */

    spamicity = (*((rf_method_t *)method)->get_spamicity)( robn, P, Q );

    if (need_stats && robn != 0)
	rstats_fini(robn, P, Q, spamicity );

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "### rob_compute_spamicity() ends\n");

    rob_stats.s.spamicity = spamicity;

    return (spamicity);
}

double rob_get_spamicity(size_t robn, FLOAT P, FLOAT Q)
{
    if (robn == 0)
	rob_stats.s.spamicity = robx;
    else
    {
	double r = 1.0 / robn;
	static double ln2 = 0.6931472;				/* ln(2) */

	rob_stats.robn = robn;

	/* convert to natural logs */
	rob_stats.p_ln = log(P.mant) + P.exp * ln2;		/* invlogsum */
	rob_stats.q_ln = log(Q.mant) + Q.exp * ln2;		/* logsum    */

	rob_stats.p_pr = 1.0 - exp(rob_stats.p_ln * r);		/* Robinson's P */
	rob_stats.q_pr = 1.0 - exp(rob_stats.q_ln * r);		/* Robinson's Q */

	rob_stats.s.spamicity = (1.0 + (rob_stats.p_pr - rob_stats.q_pr) / (rob_stats.p_pr + rob_stats.q_pr)) / 2.0;
    }

    return rob_stats.s.spamicity;
}

void rob_print_summary(void)
{
    if (!Rtable) {
	(void)fprintf(stdout, "%-*s %5lu %9.6f %9.6f %9.6f\n",
		      MAXTOKENLEN+2, "P_Q_S_invs_logs_md", 
		      (unsigned long)rob_stats.robn,
		      rob_stats.p_pr, rob_stats.q_pr, rob_stats.s.spamicity);
	(void)fprintf(stdout, "%-*s %9.2e %9.2e %6.3f\n",
		      MAXTOKENLEN+2+6, " ", robs, robx, min_dev);
    }
    else
	(void)fprintf(stdout, "%-*s %5lu %9.6f %9.6f %9.6f %9.3f %9.3f %5.3f\n",
		      MAXTOKENLEN+2, "P_Q_S_invs_logs_md", 
		      (unsigned long)rob_stats.robn,
		      rob_stats.p_pr, rob_stats.q_pr, rob_stats.s.spamicity, 
		      rob_stats.p_ln, rob_stats.q_ln, min_dev);
}

void rob_initialize_with_parameters(rob_stats_t *stats, double _min_dev, double _spam_cutoff)
{
    word_t *word_robx = word_new((const byte *)ROBX_W, strlen(ROBX_W));

    mth_initialize( stats, ROBINSON_MAX_REPEATS, _min_dev, _spam_cutoff, ROBINSON_GOOD_BIAS );

    /*
    ** If we're classifying messages, we need to compute the scalefactor 
    ** (from the .MSG_COUNT values)
    ** If we're registering tokens, we needn't get .MSG_COUNT
    */

    compute_msg_counts();
    if (fabs(robs) < EPS)
	robs = ROBS;

    if (fabs(robx) < EPS && word_list->dsh != NULL)
    {
	int ret;
	dsv_t val;

	/* Note: .ROBX is scaled by 1000000 in the wordlist */
	ret = ds_read(word_list->dsh, word_robx, &val);
	if (ret != 0)
	    robx = ROBX;
	else {
	    /* If found, unscale; else use predefined value */
	    uint l_robx = val.count[IX_SPAM];
	    robx = l_robx ? (double)l_robx / 1000000 : ROBX;
	}
    }

    if (robx < 0.0 || 1.0 < robx) {
	fprintf(stderr, "Invalid robx value (%f).  Must be between 0.0 and 1.0\n", robx);
	exit(EX_ERROR);
    }

    word_free(word_robx);

    return;
}

void rob_initialize_constants(void)
{
    rob_initialize_with_parameters(&rob_stats, ROBINSON_MIN_DEV, ROBINSON_SPAM_CUTOFF);
}

void rob_cleanup(void)
{
    rstats_cleanup();
}

/* Done */

==> bogofilter-0.16.4.tar.gz: bogofilter-0.16.4/src/robinson.c <==
/* $Id: robinson.c,v 1.70 2004/01/20 12:55:17 relson Exp $ */

/*****************************************************************************

NAME:
   robinson.c -- implements f(w) and S, or Fisher, algorithm for computing spamicity.

******************************************************************************/

#include "common.h"

#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "bogoconfig.h"
#include "bogofilter.h"
#include "collect.h"
#include "datastore.h"
#ifdef	ENABLE_DEPRECATED_CODE
#include "degen.h"
#endif
#include "msgcounts.h"
#include "prob.h"
#include "robinson.h"
#include "rstats.h"
#include "wordhash.h"
#include "wordlists.h"

extern int Rtable;

#ifdef	ENABLE_DEPRECATED_CODE
double	thresh_rtable = 0.0;		/* used in fisher.c */
#endif

static rob_stats_t  rob_stats;

#ifdef	ENABLE_DEPRECATED_CODE
const parm_desc rob_parm_table[] =	/* needed by fisher.c */
{
    { "robx",		  CP_DOUBLE,	{ (void *) &robx } },
    { "robs",		  CP_DOUBLE,	{ (void *) &robs } },
    { "thresh_rtable",	  CP_DOUBLE,	{ (void *) &thresh_rtable } },
    { NULL,		  CP_NONE,	{ (void *) NULL } },
};
#endif

/* Function Prototypes */

#ifdef	ENABLE_DEPRECATED_CODE
static void	rob_initialize_constants(void);
static double	rob_get_spamicity(size_t robn, FLOAT P, FLOAT Q);
static void	rob_print_summary(void);
#endif

/* Static Variables */

#ifdef	ENABLE_DEPRECATED_CODE
rf_method_t rf_robinson_method = {	/* needed by config.c */
    {
	"robinson",			/* const char		  *name;		*/
	rob_parm_table,	 		/* m_parm_table		  *parm_table		*/
	rob_initialize_constants,	/* m_initialize_constants *initialize_constants	*/
	rob_compute_spamicity, 		/* m_compute_spamicity	  *compute_spamicity	*/
	mth_spamicity,			/* m_spamicity		  *spamicity		*/
	mth_status,			/* m_status		  *status		*/
	rob_print_stats, 		/* m_print_bogostats	  *print_stats		*/
	rob_cleanup, 			/* m_free		  *cleanup		*/
    },
    rob_get_spamicity,			/* rf_get_spamicity	  *get_spamicity	*/
    rob_print_summary			/* rf_print_summary	  *print_summary	*/
};
#endif

/* Function Definitions */

void rob_print_stats(FILE *fp)
{
    bool unsure = unsure_stats && ((*method->status)() == RC_UNSURE) && verbose;

    fp = NULL; 	/* quench compiler warning */
    if (Rtable || unsure || verbose >= 2
#ifdef	ENABLE_DEPRECATED_CODE
	||
	rob_stats.s.spamicity > thresh_stats ||
	rob_stats.s.spamicity > thresh_rtable
#endif
	)
	rstats_print(unsure);
}

static void wordprob_add(wordcnts_t *cnts, uint count, uint bad)
{
    if (bad)
	cnts->bad += count;
    else
	cnts->good += count;
}

static double wordprob_result(wordcnts_t *cnt)
{
    double fw = calc_prob(cnt->good, cnt->bad);
    return fw;
}

static void lookup(const word_t *token, wordcnts_t *cnts)
{
    int override=0;
    wordlist_t* list;

    if (fBogotune) {
	wordprop_t *wp = wordhash_search_memory(token);
	if (wp) {
	    cnts->good = wp->cnts.good;
	    cnts->bad  = wp->cnts.bad;
	}
	return;
    }

    for (list=word_lists; list != NULL; list=list->next)
    {
	size_t i;
	dsv_t val;

	if (override > list->override)	/* if already found */
	    break;

	if (ds_read(list->dsh, token, &val) != 0)
	    continue;			/* not found */

	if (list->ignore)		/* if on ignore list */
	    return;

	override=list->override;

	for (i=0; i<COUNTOF(val.count); i++) {
	    /* Protect against negatives */
	    if ((int) val.count[i] < 0) {
		val.count[i] = 0;
		ds_write(list->dsh, token, &val);
	    }
	    wordprob_add(cnts, val.count[i], list->bad[i]);
	}

	if (DEBUG_ROBINSON(1)) {
	    fprintf(dbgout, "%2d %2d \n", (int) cnts->good, (int) cnts->bad);
	    word_puts(token, 0, dbgout);
	    fputc('\n', dbgout);
	}
    }

    return;
}

double lookup_and_score(const word_t *token, wordcnts_t *cnts)
{
    double prob;

#ifdef	ENABLE_DEPRECATED_CODE
    const char *colon;
#endif

    if (cnts->bad == 0 && cnts->good == 0)
	lookup(token, cnts);

#ifdef	ENABLE_DEPRECATED_CODE
    if (header_degen && cnts->bad == 0 && cnts->good == 0 && (colon = strchr((const char *)(token->text), ':')) != NULL) {
	word_t *tword = word_new((unsigned const char *)(colon+1), strlen(colon+1));
	wordcnts_t tcnts;
	wordcnts_init(&tcnts);
	lookup_and_score(tword, &tcnts);
	wordcnts_incr(cnts, &tcnts);
	prob = wordprob_result(cnts);
	word_free(tword);
    }
    else
    if (cnts->bad != 0 || cnts->good != 0) {
	prob = wordprob_result(cnts);
    }
    else
    if (degen_enabled)
    {
	degen_enabled = false;	/* Disable further recursion */
	prob = degen(token, cnts);
	degen_enabled = true;	/* Enable further recursion */
    }
    else
#endif
	prob = wordprob_result(cnts);

    return prob;
}

static double compute_probability(const word_t *token, wordcnts_t *cnts)
{
    double prob;

    if (cnts->bad != 0 || cnts->good != 0 || msg_count_file)
	/* A msg-count file already has the values needed */
	prob = wordprob_result(cnts);
    else
	/* Otherwise lookup the word and get its score */
	prob = lookup_and_score(token, cnts);

    return prob;
}

double rob_compute_spamicity(wordhash_t *wh, FILE *fp) /*@globals errno@*/
/* selects the best spam/non-spam indicators and calculates Robinson's S */
{
    hashnode_t *node;

    FLOAT P = {1.0, 0};		/* Robinson's P */
    FLOAT Q = {1.0, 0};		/* Robinson's Q */

    double spamicity;
    size_t robn = 0;
    size_t count = 0;
    bool need_stats;

    (void) fp; 	/* quench compiler warning */

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "### rob_compute_spamicity() begins\n");

    Rtable |= verbose > 3;
    need_stats = Rtable || verbose || passthrough;

    if (need_stats)
	rstats_init();

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "min_dev: %f, robs: %f, robx: %f\n", 
				   min_dev, robs, robx);

    wordhash_sort(wh);

    for(node = wordhash_first(wh); node != NULL; node = wordhash_next(wh))
    {
	double prob;
	word_t *token;
	wordcnts_t *cnts;
	wordprop_t *props;

	if (!fBogotune) {
	    props = (wordprop_t *) node->buf;
	    cnts  = &props->cnts;
	    token = node->key;
	}
	else {
	    cnts = (wordcnts_t *) node;
	    token = NULL;
	}

	count += 1;

	prob = compute_probability(token, cnts);

	if (need_stats)
	    rstats_add(token, prob, cnts);

	/* Robinson's P and Q; accumulation step */
        /*
	 * P = 1 - ((1-p1)*(1-p2)*...*(1-pn))^(1/n)     [spamminess]
         * Q = 1 - (p1*p2*...*pn)^(1/n)                 [non-spamminess]
	 */
        if (fabs(EVEN_ODDS - prob) - min_dev >= EPS) {
	    int e;

	    P.mant *= 1-prob;
	    if (P.mant < 1.0e-200) {
		P.mant = frexp(P.mant, &e);
		P.exp += e;
	    }

	    Q.mant *= prob;
	    if (Q.mant < 1.0e-200) {
		Q.mant = frexp(Q.mant, &e);
		Q.exp += e;
	    }
            robn ++;
        }

	if (DEBUG_ROBINSON(3)) {
	    (void)fprintf(dbgout, "%3lu %3lu %f ",
			  (unsigned long)robn, (unsigned long)count, prob);
	    (void)word_puts(token, 0, dbgout);
	    (void)fputc('\n', dbgout);
	}
    }

    /* Robinson's P, Q and S
    ** S = (P - Q) / (P + Q)                        [combined indicator]
    */

    spamicity = (*((rf_method_t *)method)->get_spamicity)( robn, P, Q );

    if (need_stats && robn != 0)
	rstats_fini(robn, P, Q, spamicity );

    if (DEBUG_ROBINSON(2)) fprintf(dbgout, "### rob_compute_spamicity() ends\n");

    rob_stats.s.spamicity = spamicity;

    return (spamicity);
}

#ifdef	ENABLE_DEPRECATED_CODE
double rob_get_spamicity(size_t robn, FLOAT P, FLOAT Q)
{
    if (robn == 0)
	rob_stats.s.spamicity = robx;
    else
    {
	double r = 1.0 / robn;
	static double ln2 = 0.6931472;				/* ln(2) */

	rob_stats.robn = robn;

	/* convert to natural logs */
	rob_stats.p_ln = log(P.mant) + P.exp * ln2;		/* invlogsum */
	rob_stats.q_ln = log(Q.mant) + Q.exp * ln2;		/* logsum    */

	rob_stats.p_pr = 1.0 - exp(rob_stats.p_ln * r);		/* Robinson's P */
	rob_stats.q_pr = 1.0 - exp(rob_stats.q_ln * r);		/* Robinson's Q */

	rob_stats.s.spamicity = (1.0 + (rob_stats.p_pr - rob_stats.q_pr) / (rob_stats.p_pr + rob_stats.q_pr)) / 2.0;
    }

    return rob_stats.s.spamicity;
}
#endif

#ifdef	ENABLE_DEPRECATED_CODE
void rob_print_summary(void)
{
    if (!Rtable) {
	(void)fprintf(stdout, "%-*s %5lu %9.6f %9.6f %9.6f\n",
		      MAXTOKENLEN+2, "P_Q_S_invs_logs_md", 
		      (unsigned long)rob_stats.robn,
		      rob_stats.p_pr, rob_stats.q_pr, rob_stats.s.spamicity);
	(void)fprintf(stdout, "%-*s %9.2e %9.2e %6.3f\n",
		      MAXTOKENLEN+2+6, " ", robs, robx, min_dev);
    }
    else
	(void)fprintf(stdout, "%-*s %5lu %9.6f %9.6f %9.6f %9.3f %9.3f %5.3f\n",
		      MAXTOKENLEN+2, "P_Q_S_invs_logs_md", 
		      (unsigned long)rob_stats.robn,
		      rob_stats.p_pr, rob_stats.q_pr, rob_stats.s.spamicity, 
		      rob_stats.p_ln, rob_stats.q_ln, min_dev);
}
#endif

void rob_initialize_with_parameters(rob_stats_t *stats, double _min_dev, double _spam_cutoff)
{
    word_t *word_robx = word_new((const byte *)ROBX_W, strlen(ROBX_W));

    mth_initialize( stats, ROBINSON_MAX_REPEATS, _min_dev, _spam_cutoff, ROBINSON_GOOD_BIAS );

    /*
    ** If we're classifying messages, we need to compute the scalefactor 
    ** (from the .MSG_COUNT values)
    ** If we're registering tokens, we needn't get .MSG_COUNT
    */

    compute_msg_counts();
    if (fabs(robs) < EPS)
	robs = ROBS;

    if (fabs(robx) < EPS && word_list->dsh != NULL)
    {
	int ret;
	dsv_t val;

	/* Note: .ROBX is scaled by 1000000 in the wordlist */
	ret = ds_read(word_list->dsh, word_robx, &val);
	if (ret != 0)
	    robx = ROBX;
	else {
	    /* If found, unscale; else use predefined value */
	    uint l_robx = val.count[IX_SPAM];
	    robx = l_robx ? (double)l_robx / 1000000 : ROBX;
	}
    }

    if (robx < 0.0 || 1.0 < robx) {
	fprintf(stderr, "Invalid robx value (%f).  Must be between 0.0 and 1.0\n", robx);
	exit(EX_ERROR);
    }

    word_free(word_robx);

    return;
}

#ifdef	ENABLE_DEPRECATED_CODE
void rob_initialize_constants(void)
{
    rob_initialize_with_parameters(&rob_stats, ROBINSON_MIN_DEV, ROBINSON_SPAM_CUTOFF);
}
#endif

void rob_cleanup(void)
{
    rstats_cleanup();
}

/* Done */
